import sys
import re
import subprocess

from markdown.preprocessors import Preprocessor
from markdown.extensions import Extension as BaseExtension

from pymdgen import doc_module


class GenCodeDocs(Preprocessor):
    """
    output code docs for specified python module
    """

    regex = re.compile(r"^\{pymdgen:(.+)\}$")

    def run(self, lines):
        new_lines = []
        for line in lines:
            m = self.regex.match(line)
            if m:
                new_lines.extend(self.generate(m.group(1)))
            else:
                new_lines.append(line)
        return new_lines

    def generate(self, module_name):
        return doc_module(module_name, section_level=1)


class GenCommandOutput(Preprocessor):
    """
    render command output
    """

    regex = re.compile(r"^\{pymdgen-cmd:(.+)\}$")

    def run(self, lines):
        new_lines = []
        for line in lines:
            m = self.regex.match(line)
            if m:
                new_lines.extend(self.generate(m.group(1)))
            else:
                new_lines.append(line)
        return new_lines

    def generate(self, command):
        v = sys.version_info[0]
        output = subprocess.check_output(command.split(" "))
        lines = ["```"]
        for line in output.split(b"\n"):
            if v == 2:
                lines.append(u"{}".format(line).strip("'"))
            else:
                lines.append(u"{}".format(line)[1:].strip("'"))

        lines.append("```")
        return lines


class Extension(BaseExtension):
    def extendMarkdown(self, md):
        md.preprocessors.register(GenCodeDocs(md), "pymdgencode", 175)
        md.preprocessors.register(GenCommandOutput(md), "pymdgencmd", 175)



import importlib
import inspect
import logging
import re

log = logging.getLogger("pymdgen")


def adjust_header_indent(line, section_level):
    """
    Parses a line for a markdown header and fixes the
    indent for the specified section level

    eg. '## Header' with section_level=1 will become
    '# Header"

    Arguments:

    - line(str)
    - section_level(int)

    Returns:

    - str
    """

    m = re.match("[#+] (.+)", line)
    if m:
        line = u"{} {}".format("#" * (section_level), m.group(1))
    return line


def doc_func(name, func, section_level=4):

    """
    return markdown formatted documentation for a function

    Arguments:

    - name(str): function name #FIXME: why is this manual?
    - func(function)
    - section_level(int): markdown section level
    """

    is_property = False

    if isinstance(func, property):
        func = func.fget
        is_property = True

    output = []
    docstr = inspect.getdoc(func)
    # skip functions without a docstr
    if not docstr:
        return output

    spec = inspect.getargspec(func)
    display = []
    end_args = []

    # *args and **kwargs
    if spec[1]:
        end_args.append("*" + spec[1])
    if spec[2]:
        end_args.append("**" + spec[2])

    # check for args with defaults
    if spec[3]:
        args = spec[0][-len(spec[3]) :]
        default_args = list(zip(args, spec[3]))

        # set args to rest
        args = spec[0][: -len(spec[3])]
    else:
        args = spec[0]
        default_args = []

    if args:
        display.append(", ".join(args))
    if default_args:
        display.append(", ".join("%s=%s" % x for x in default_args))
    if end_args:
        display.append(", ".join(end_args))

    if name.find("__") == 0:
        title = "\{}".format(name)
    else:
        title = name
    title = title + " `(" + ", ".join(display) + ")`"

    output.append("{} {}".format("#" * section_level, title))
    output.append("")
    output.append(docstr)
    output.append("")
    output.append("---")

    return output


def parse_class_docstr(docstr, list_class_attributes, section_level):

    """
    Takes a class doc string markdown and parses it line for line
    to fix and header indent issues as well as collect instanced
    attribute docs

    Arguments

    - docstr(str): markdown formatted class docstr
    - list_class_attributes(list): collect instance attribute docs into
      this list
    - section_level(int): header indent level

    Returns:

    - list: parsed/fixed output split into lines
    """

    attributes_regex = "[#+] Instanced (Attributes|Properties)"
    header_regex = "[#+] (.+)"

    collect_attributes = False
    out = []

    for line in docstr.split("\n"):
        if re.match(attributes_regex, line):
            collect_attributes = True
        elif collect_attributes and line.strip():
            list_class_attributes.append(line)
        elif collect_attributes and list_class_attributes and not line.strip():
            collect_attributes = False
        else:
            out.append(adjust_header_indent(line, section_level + 1))

    return out


def doc_class(name, cls, section_level=3):
    """
    return markdown formatted documentation for a class

    Arguments

    - name(str): function name #FIXME: why is this manual?
    - cls(class)
    - section_level(int): markdown section level
    """

    output = []
    docstr = inspect.getdoc(cls)
    # skip functions without a docstr
    if not docstr:
        return output

    head_indent = "#" * (section_level + 1)

    out_class_attributes = ["", "{} Class Attributes".format(head_indent), ""]
    out_instanced_attributes = [
        "",
        "{} Instanced Attributes".format(head_indent),
        "",
        "These attributes / properties will be available on instances of the class",
        "",
    ]

    out_class_methods = ["", "{} Class Methods".format(head_indent), ""]
    out_methods = ["", "{} Methods".format(head_indent), ""]

    list_instanced_attributes = []
    list_class_attributes = []
    list_class_methods = []
    list_methods = []

    # parse the class docstr markdown to:
    # - fix header indent according to section level
    # - collect arbitrary instance attribute documentation
    out_docstr = parse_class_docstr(docstr, list_instanced_attributes, section_level)

    # full mro is probably overkill?
    # base_classes = inspect.getmro(cls)
    base_classes = cls.__bases__
    base_classes = (c.__module__ + "." + c.__name__ for c in base_classes)

    output.append("{} {}".format("#" * section_level, name))
    output.append("")
    output.append("```")
    output.append(name + "(" + ", ".join(base_classes) + ")")
    output.append("```")
    output.append("")
    output.extend(out_docstr)
    output.append("")

    functions = sorted(list(cls.__dict__.items()), key=lambda x: x[0])

    for func_name, func in functions:
        if inspect.isfunction(func):
            list_methods.extend(doc_func(func_name, func, section_level + 2))
        elif isinstance(func, classmethod):
            list_class_methods.extend(
                doc_func(func_name, func.__func__, section_level + 2)
            )
        elif isinstance(func, property):
            list_instanced_attributes.extend(doc_property(func_name, func))
        elif hasattr(func, "help"):
            list_class_attributes.extend(doc_attribute(func_name, func))

    # Append class attribute documentation to output

    if list_class_attributes:
        output.extend(out_class_attributes)
        output.extend(sorted(list_class_attributes))

    # Append instanced attributed documentation to output

    if list_instanced_attributes:
        output.extend(out_instanced_attributes)
        output.extend(sorted(list_instanced_attributes))

    # Appemd class method documentation to output

    if list_class_methods:
        output.extend(out_class_methods)
        output.extend(list_class_methods)

    # Append method documentation to output

    if list_methods:
        output.extend(out_methods)
        output.extend(list_methods)

    output.append("")

    return output


def doc_attribute(name, attribute):
    """
    return markdown formatted documentation for a class attribute

    This is an experimental feature that will document any attribute
    set on class as long as the attribute it self has `help` property

    Argument(s):

    - name(str)
    - attribute(function|class|instance)

    Returns:

    - list
    """
    if not hasattr(attribute, "help"):
        return

    output = []
    type_name = None

    if inspect.isclass(attribute):
        type_name = "`{} Class`".format(attribute.__name__)
    elif hasattr(attribute, "__class__"):
        type_name = "`{} Instance`".format(attribute.__class__.__name__)

    try:
        output.append("- {} ({}): {}".format(name, type_name, attribute.help))
    except:
        pass

    return output


def doc_property(name, prop):
    """
    return markdown formatted documentation for a class property

    Argument(s):

    - name(str)
    - prop(property)

    Returns:

    - list
    """
    docstr = inspect.getdoc(prop.fget)
    return ["- {} (`{}`): {}".format(name, "@property", docstr)]


def doc_module(name, debug=False, section_level=3):

    """
    return markdown formatted documentation for a module

    Arguments:

    - name(str): module name
    - debug(bool): log debug messages
    - section_level(int): markdown section level
    """

    if "/" in name or name.endswith(".py"):
        name = name.replace("/", ".")
        name = name.rstrip(".py")

    module = importlib.import_module(name)
    output = []

    output.append("{} {}".format("#" * section_level, module.__name__))

    docstr = inspect.getdoc(module)
    if docstr:
        output.extend(["", docstr])

    output.append("")

    for k, v in inspect.getmembers(module):
        if k == "__builtins__":
            continue
        log.debug("checking %s:%s" % (v, k))
        if inspect.isfunction(v):
            if v.__module__ == module.__name__:
                output.extend(doc_func(k, v, section_level + 1))
        if inspect.isclass(v):
            if v.__module__ == module.__name__:
                output.extend(doc_class(k, v, section_level + 1))

    return output



#!/bin/env python

import argparse
import importlib
import inspect
import logging

from pymdgen import doc_module


def run(modules, debug, section_level):

    if debug:
        logging.basicConfig(level=logging.DEBUG)

    output = []
    for name in modules:
        md = doc_module(name, debug=debug, section_level=section_level)
        for line in md:
            output.append(line)
    return output


def main():
    parser = argparse.ArgumentParser(
        description="Inspects given python modules and prints markdown"
    )

    parser.add_argument(
        "--debug", dest="debug", action="store_true", help="display debug messages"
    )
    parser.add_argument(
        "--section-level", type=int, default=3, help="markdown section lavel"
    )
    parser.add_argument("modules", nargs="+")

    args = parser.parse_args()

    debug = args.debug
    modules = args.modules
    section_level = args.section_level

    for line in run(modules, debug, section_level):
        print(line)


if __name__ == "__main__":
    main()



"""
A module to use as a target during unit tests
"""


class prop(object):
    help = "property help"


class dummy(object):
    """
    this is a dummy class

    # Instanced Properties

    - arbitrary property
    - another arbitrary property

    more class docstr
    """

    test_property = prop()

    @classmethod
    def dummy_classmethod(self):
        """ this is a dummy classmethod """

        return

    def dummy_method(self):
        """ this is a dummy func """

        return

    @property
    def dummy_property(self):
        """
        this is a dummy property
        """

        return "dummy"


def dummy_func():
    """ this is a dummy func """
    return
